__author__ = 'breddels'
import numpy as np
import logging
import threading
import uuid
import time
import ast

try:
	import __builtin__
except ImportError:
	import builtins as __builtin__

from .dataset import Dataset, Task, _issequence
from .legacy import Subspace
import vaex.promise
import vaex.settings
import vaex.utils
try:
	import Cookie  # py2
except ImportError:
	import http.cookies as Cookie  # py3

from tornado.httpclient import AsyncHTTPClient, HTTPClient
import tornado.httputil
import tornado.websocket
from tornado.concurrent import Future
from tornado import gen
from .dataset import default_shape

logger = logging.getLogger("vaex.remote")

DEFAULT_REQUEST_TIMEOUT = 60 * 5 # 5 minutes

def wrap_future_with_promise(future):
	if isinstance(future, vaex.promise.Promise): # TODO: not so nice, sometimes we pass a promise
		return future
	promise = vaex.promise.Promise()
	def callback(future):
		#print(("callback", future, future.result()))
		e = future.exception()
		if e:
			#print(("reject", e))
			promise.reject(e)
		else:
			promise.fulfill(future.result())
	future.add_done_callback(callback)
	return promise

import tornado.ioloop

import threading
import json

def listify(value):
	if isinstance(value, list):
		value = list([listify(item) for item in value])
	if hasattr(value, "tolist"):
		value = value.tolist()
	return value


try:
	from urllib.request import urlopen
	from urllib.parse import urlparse, urlencode
except ImportError:
	from urlparse import urlparse
	from urllib import urlopen, urlencode

import threading

def _check_error(object):
	if "error" in object:
		raise RuntimeError("Server responded with error: %r" % object["error"])


class ServerRest(object):
	def __init__(self, hostname, port=5000, base_path="/", background=False, thread_mover=None, websocket=True):
		self.hostname = hostname
		self.port = port
		self.base_path = base_path if base_path.endswith("/") else (base_path + "/")
		#if delay:
		event = threading.Event()
		self.thread_mover = thread_mover or (lambda fn, *args, **kwargs: fn(*args, **kwargs))
		logger.debug("thread mover: %r", self.thread_mover)

		# jobs maps from uid to tasks
		self.jobs = {}

		def ioloop_threaded():
			logger.debug("creating tornado io_loop")
			self.io_loop = tornado.ioloop.IOLoop().current()
			event.set()
			logger.debug("started tornado io_loop...")

			self.io_loop.start()
			self.io_loop.close()
			logger.debug("stopped tornado io_loop")

		io_loop = tornado.ioloop.IOLoop.current(instance=False)
		if True:# io_loop:# is None:
			logger.debug("no current io loop, starting it in thread")
			thread = threading.Thread(target=ioloop_threaded)
			thread.setDaemon(True)
			thread.start()
			event.wait()
		else:
			logger.debug("using current io loop")
			self.io_loop = io_loop

		self.io_loop.make_current()
		#if async:
		self.http_client_async = AsyncHTTPClient()
		self.http_client = HTTPClient()
		self.user_id = vaex.settings.webclient.get("cookie.user_id")
		self.use_websocket = websocket
		self.websocket = None

		self.submit = self.submit_http

		if self.use_websocket:
			logger.debug("connect via websocket")
			self.submit = self.submit_websocket
			self._websocket_connect()
			logger.debug("websocket connected")

	def close(self):
		#self.http_client_async.
		self.http_client.close()
		if self.use_websocket:
			if self.websocket:
					self.websocket.close()
		self.io_loop.stop()


	def _websocket_connect(self):
		def connected(websocket):
			logger.debug("connected to websocket: %s" % self._build_url(""))
		def failed(reason):
			logger.error("failed to connect to %s" % self._build_url(""))
		self.websocket_connected = vaex.promise.Promise()
		self.websocket_connected.then(connected, failed)
		if 0:
			connected = wrap_future_with_promise(tornado.websocket.websocket_connect(self._build_url("websocket"), on_message_callback=self._on_websocket_message))
			connected.get()
			self.websocket_connected.fulfill(connected)
		def do():
			try:
				logger.debug("wrapping promise")
				connected = wrap_future_with_promise(tornado.websocket.websocket_connect(self._build_url("websocket"), on_message_callback=self._on_websocket_message))
				logger.debug("continue")
				self.websocket_connected.fulfill(connected)
			except:
				logger.exception("error connecting")
				#raise
		logger.debug("add callback")
		self.io_loop.add_callback(do)
		logger.debug("added callback: ")
		#self.io_loop.start()
		#if self.port == 29345:
		#	import pdb
		#	pdb.set_trace()
		logger.debug("waiting for connection")
		result = self.websocket_connected.get()
		logger.debug("websocket connected")
		if self.websocket_connected.isRejected:
			raise self.websocket.reason




	def _on_websocket_message(self, msg):
		json_data, data = msg.split(b"\n", 1)
		response = json.loads(json_data.decode("utf8"))
		if data:
			import zlib
			data = zlib.decompress(data)
			numpy_array = np.fromstring(data, dtype=np.dtype(response["dtype"])).reshape(ast.literal_eval(response["shape"]))
			response["result"] = numpy_array
		import sys
		if sys.getsizeof(msg) > 1024*4:
			logger.debug("socket read message: <large amount of data>",)
		else:
			logger.debug("socket read message: %s", msg)
			logger.debug("json response: %r", response)
		# for the moment, job == task, in the future a job can be multiple tasks
		job_id = response.get("job_id")
		if job_id:
			try:
				phase = response["job_phase"]
				logger.debug("job update %r, phase=%r", job_id, phase)
				if phase == "COMPLETED":
					result = response["result"]#[0]
					#logger.debug("completed job %r, result=%r", job_id, result)
					task = self.jobs[job_id]
					logger.debug("completed job %r (delay=%r, thread_mover=%r)", job_id, task.delay, self.thread_mover)
					processed_result = task.post_process(result)
					if task.delay:
						self.thread_mover(task.fulfill, processed_result)
					else:
						task.fulfill(processed_result)
				elif phase == "EXCEPTION":
					logger.error("exception happened at server side: %r", response)
					class_name = response["exception"]["class"]
					msg = response["exception"]["msg"]
					exception = getattr(__builtin__, class_name)(msg)
					logger.debug("error in job %r, exception=%r", job_id, exception)
					task = self.jobs[job_id]
					if task.delay:
						self.thread_mover(task.reject, exception)
					else:
						task.reject(exception)
				elif phase == "ERROR":
					logger.error("error happened at server side: %r", response)
					msg = response["error"]
					exception = RuntimeError("error at server: %r" % msg)
					task = self.jobs[job_id]
					if task.delay:
						self.thread_mover(task.reject, exception)
					else:
						task.reject(exception)
				elif phase == "PENDING":
					fraction = response["progress"]
					logger.debug("pending?: %r", phase)
					task = self.jobs[job_id]
					if task.delay:
						self.thread_mover(task.signal_progress.emit, fraction)
					else:
						task.signal_progress.emit(fraction)
			except Exception as e:
				logger.exception("error in handling job", e)
				task = self.jobs[job_id]
				if task.delay:
					self.thread_mover(task.reject, e)
				else:
					task.reject(e)


	def wait(self):
		io_loop = tornado.ioloop.IOLoop.instance()
		io_loop.start()

	def submit_websocket(self, path, arguments, delay=False, progress=None, post_process=lambda x: x):
		assert self.use_websocket

		task = TaskServer(post_process=post_process, delay=delay)
		progressbars = vaex.utils.progressbars(progress)
		progressbars.add_task(task)
		logger.debug("created task: %r, %r (delay=%r)" % (path, arguments, delay))
		job_id = str(uuid.uuid4())
		self.jobs[job_id] = task
		arguments["job_id"] = job_id
		arguments["path"] = path
		arguments["user_id"] = self.user_id
		#arguments = dict({key: (value.tolist() if hasattr(value, "tolist") else value) for key, value in arguments.items()})
		arguments = dict({key: listify(value) for key, value in arguments.items()})

		def do():
			def write(socket):
				try:
					logger.debug("write to websocket: %r", arguments)
					socket.write_message(json.dumps(arguments))
					return
				except:
					import traceback
					traceback.print_exc()
			#return
			logger.debug("will schedule a write to the websocket")
			self.websocket_connected.then(write).end()#.then(task.fulfill)

		self.io_loop.add_callback(do)
		logger.debug("we can continue (main thread is %r)", threading.currentThread())
		if delay:
			return task
		else:
			return task.get()

	def submit_http(self, path, arguments, post_process, delay, progress=None, **kwargs):
		def pre_post_process(response):
			cookie = Cookie.SimpleCookie()
			for cookieset in response.headers.get_list("Set-Cookie"):
				cookie.load(cookieset)
				logger.debug("cookie load: %r", cookieset)
			logger.debug("cookie: %r", cookie)
			if "user_id" in cookie:
				user_id = cookie["user_id"].value
				logger.debug("user_id: %s", user_id)
				if self.user_id != user_id:
					self.user_id = user_id
					vaex.settings.webclient.store("cookie.user_id", self.user_id)
			data = response.body
			is_json = False
			logger.info("response is: %r", response.body)
			logger.info("content_type is: %r", response.headers["Content-Type"])
			if response.headers["Content-Type"] == "application/numpy-array":
				shape, dtype, data = response.body.split(b"\n", 2)
				shape = shape.decode("ascii")
				dtype = dtype.decode("ascii")
				import ast
				numpy_array = np.fromstring(data, dtype=np.dtype(dtype)).reshape(ast.literal_eval(shape))
				return post_process(numpy_array)
			else:
				try:
					data = json.loads(response.body.decode("ascii"))
					is_json = True
				except Exception as e:
					logger.info("couldn't convert to json (error is %s, assume it's raw data): %s", e, data)
					#logger.info("couldn't convert to json (error is %s, assume it's raw data)", e)
				if is_json:
					self._check_exception(data)
					return post_process(data["result"])
				else:
					return post_process(data)

		arguments = {key: listify(value) for key, value in arguments.items()}
		import pdb
		arguments_json = {key:json.dumps(value) for key, value in arguments.items()}
		headers = tornado.httputil.HTTPHeaders()

		url = self._build_url(path +"?" + urlencode(arguments_json))
		logger.debug("fetch %s, delay=%r", url, delay)

		if self.user_id is not None:
			headers.add("Cookie", "user_id=%s" % self.user_id)
			logger.debug("adding user_id %s to request", self.user_id)
		if delay:
			task = TaskServer(pre_post_process, delay=delay)
			# tornado doesn't like that we call fetch while ioloop is running in another thread, we should use ioloop.add_callbacl
			def do():
				self.thread_mover(task.signal_progress.emit, 0.5)
				future = self.http_client_delay.fetch(url, headers=headers, request_timeout=DEFAULT_REQUEST_TIMEOUT, **kwargs)
				def thread_save_succes(value):
					self.thread_mover(task.signal_progress.emit, 1.0)
					self.thread_mover(task.fulfill, value)
				def thread_save_failure(value):
					self.thread_mover(task.reject, value)
				wrap_future_with_promise(future).then(pre_post_process).then(thread_save_succes, thread_save_failure)
			self.io_loop.add_callback(do)
			return task #promise.then(self._move_to_thread)
		else:
			return pre_post_process(self.http_client.fetch(url, headers=headers, request_timeout=DEFAULT_REQUEST_TIMEOUT, **kwargs))

	def datasets(self, as_dict=False, delay=False):
		def post(result):
			logger.debug("datasets result: %r", result)
			def create(server, state):
				dataset = DatasetRest(self, name=state['name'],
							          length_original=state['length_original'],
							          column_names=state['column_names'],
							          dtypes=state['dtypes'])
				dataset.state_set(state['state'])
				return dataset
			datasets = [create(self, kwargs) for kwargs in result]
			logger.debug("datasets: %r", datasets)
			return datasets if not as_dict else dict([(ds.name, ds) for ds in datasets])
		return self.submit(path="datasets", arguments={}, post_process=post, delay=delay)

	def _build_url(self, method):
		protocol = "ws" if self.use_websocket else "http"
		return "%s://%s:%d%s%s" % (protocol, self.hostname, self.port, self.base_path, method)

	def _call_subspace(self, method_name, subspace, **kwargs):
		def post_process(result):
			if method_name == "histogram": # histogram is the exception..
				# TODO: don't do binary transfer, just json, now we cannot handle exception
				import base64
				#logger.debug("result: %r", result)
				#result = base64.b64decode(result) #.decode("base64")
				#result = base64.
				data = np.fromstring(result, dtype=np.float64)
				shape = (kwargs["size"],) * subspace.dimension
				data = data.reshape(shape)
				return data
			else:
				try:
					return np.array(result)
				except ValueError:
					return result
		dataset_name = subspace.dataset.name
		expressions = subspace.expressions
		delay = subspace.delay
		path = "datasets/%s/%s" % (dataset_name, method_name)
		url = self._build_url(path)
		arguments = dict(kwargs)
		dataset_remote = subspace.dataset
		arguments["selection"] = subspace.is_masked
		arguments['state'] = dataset_remote.state_get()
		arguments['auto_fraction'] = dataset_remote.get_auto_fraction()
		arguments.update(dict(expressions=expressions))
		return self.submit(path, arguments, post_process=post_process, delay=delay)

	def _call_dataset(self, method_name, dataset_remote, delay, numpy=False, progress=None, **kwargs):
		def post_process(result):
			#result = self._check_exception(json.loads(result.body))["result"]
			if numpy:
				return np.fromstring(result, dtype=np.float64)
			else:
				try:
					return np.array(result)
				except ValueError:
					return result
		path = "datasets/%s/%s" % (dataset_remote.name, method_name)
		arguments = dict(kwargs)
		arguments['state'] = dataset_remote.state_get()
		arguments['auto_fraction'] = dataset_remote.get_auto_fraction()
		body = urlencode(arguments)

		return self.submit(path, arguments, post_process=post_process, progress=progress, delay=delay)

	def _schedule_call(self, method_name, dataset_remote, delay, **kwargs):
		def post_process(result):
			#result = self._check_exception(json.loads(result.body))["result"]
			try:
				return np.array(result)
			except ValueError:
				return result
		method = "%s/%s" % (dataset_remote.name, method_name)
		return self.schedule(path, arguments, post_process=post_process, delay=delay)

	def _check_exception(self, reply_json):
		if "exception" in reply_json:
			logger.error("exception happened at server side: %r", reply_json)
			class_name = reply_json["exception"]["class"]
			msg = reply_json["exception"]["msg"]
			raise getattr(__builtin__, class_name)(msg)
		if "error" in reply_json:
			raise ValueError("unknown error occured at server")
		else:
			return reply_json

class SubspaceRemote(Subspace):
	def toarray(self, list):
		return np.array(list)

	@property
	def dimension(self):
		return len(self.expressions)

	def _task(self, promise):
		"""Helper function for returning tasks results, result when immediate is True, otherwise the task itself, which is a promise"""
		if self.delay:
			return promise
		else:
			return promise

	def sleep(self, seconds, delay=False):
		return self.dataset.server.call("sleep", seconds, delay=delay)

	def minmax(self):
		return self._task(self.dataset.server._call_subspace("minmax", self))
		#return self._task(task)

	def histogram(self, limits, size=256, weight=None):
		return self._task(self.dataset.server._call_subspace("histogram", self, size=size, limits=limits, weight=weight))

	def nearest(self, point, metric=None):
		point = vaex.utils.make_list(point)
		result = self.dataset.server._call_subspace("nearest", self, point=point, metric=metric)
		return self._task(result)

	def mean(self):
		return self.dataset.server._call_subspace("mean", self)

	def correlation(self, means=None, vars=None):
		return self.dataset.server._call_subspace("correlation", self, means=means, vars=vars)

	def var(self, means=None):
		return self.dataset.server._call_subspace("var", self, means=means)

	def sum(self):
		return self.dataset.server._call_subspace("sum", self)

	def limits_sigma(self, sigmas=3, square=False):
		return self.dataset.server._call_subspace("limits_sigma", self, sigmas=sigmas, square=square)

	def mutual_information(self, limits=None, size=256):
		return self.dataset.server._call_subspace("mutual_information", self, limits=limits, size=size)

class DatasetRemote(Dataset):
	def __init__(self, name, server, column_names):
		super(DatasetRemote, self).__init__(name, column_names)
		self.server = server

import astropy.units

class DatasetRest(DatasetRemote):
	def __init__(self, server, name, column_names, dtypes, length_original):
		DatasetRemote.__init__(self, name, server.hostname, column_names)
		self.server = server
		self.name = name
		self.column_names = column_names
		self.dtypes = {name: np.dtype(dtype) for name, dtype in dtypes.items()}
		for column_name in self.get_column_names(virtual=True, strings=True):
			self._save_assign_expression(column_name)
		self._length_original = length_original
		self._length_unfiltered = length_original
		self._index_end = length_original
		self.path = self.filename = self.server._build_url("%s" % name)
		self.fraction = 1
		self.executor = ServerExecutor()

	def copy(self, column_names=None, virtual=True):
		dtypes = {name:self.dtype(name) for name in self.get_column_names(strings=True, virtual=False)}
		ds = DatasetRest(self.server, self.name, self.column_names, dtypes=dtypes, length_original=self._length_original)
		state = self.state_get()
		if not virtual:
			state['virtual_columns'] = {}
		ds.state_set(state)
		return ds

	def count(self, expression=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False, edges=False, progress=None):
		return self._delay(delay, self.server._call_dataset("count", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection, edges=edges))

	def mean (self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
		return self._delay(delay, self.server._call_dataset("mean", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

	def sum(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
		return self._delay(delay, self.server._call_dataset("sum", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

	def var(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
		return self._delay(delay, self.server._call_dataset("var", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

	def minmax(self, expression, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
		return self._delay(delay, self.server._call_dataset("minmax", self, delay=True, progress=progress, expression=expression, binby=binby, limits=limits, shape=shape, selection=selection))

	#def count(self, expression=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False):
	def cov        (self, x, y=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
		return self._delay(delay, self.server._call_dataset("cov", self, delay=True, progress=progress, x=x, y=y, binby=binby, limits=limits, shape=shape, selection=selection))

	def correlation(self, x, y=None, binby=[], limits=None, shape=default_shape, sort=False, sort_key=np.abs, selection=False, delay=False, progress=None):
		# TODO: sort and sort_key should be done locally
		return self._delay(delay, self.server._call_dataset("correlation", self, delay=True, progress=progress, x=x, y=y, binby=binby, limits=limits, shape=shape, selection=selection))

	def _delay(self, delay, task, progressbar=False):
		if delay:
			return task
		else:
			result = task.get()
			logger.debug("result = %r", result)
			return result

	def dtype(self, expression):
		if expression in self.dtypes:
			return self.dtypes[expression]
		else:
			return np.zeros(1, dtype=np.float64).dtype

	def is_local(self): return False

	def __repr__(self):
		name = self.__class__.__module__ + "." +self.__class__.__name__
		return "<%s(server=%r, name=%r, column_names=%r, __len__=%r)> instance at 0x%x" % (name, self.server, self.name, self.column_names, len(self), id(self))

	def __call__(self, *expressions, **kwargs):
		return SubspaceRemote(self, expressions, kwargs.get("executor") or self.executor, delay=kwargs.get("delay", False))

	def evaluate(self, expression, i1=None, i2=None, out=None, selection=None, delay=False):
		expression = vaex.dataset._ensure_strings_from_expressions(expression)
		"""basic support for evaluate at server, at least to run some unittest, do not expect this to work from strings"""
		result = self.server._call_dataset("evaluate", self, expression=expression, i1=i1, i2=i2, selection=selection, delay=delay)
		# TODO: we ignore out
		return result


# we may get rid of this when we group together tasks
class ServerExecutor(object):
	def __init__(self):
		self.signal_begin = vaex.events.Signal("begin")
		self.signal_progress = vaex.events.Signal("progress")
		self.signal_end = vaex.events.Signal("end")
		self.signal_cancel = vaex.events.Signal("cancel")

	def execute(self):
		logger.debug("dummy execute")

from vaex.dataset import Task
class TaskServer(Task):
	def __init__(self, post_process, delay):
		vaex.dataset.Task.__init__(self, None, [])
		self.post_process = post_process
		self.delay = delay
		self.task_queue = []

	def schedule(self, task):
		self.task_queue.append(task)
		logger.info("task added, queue: %r", self.task_queue)
		return task


	def execute(self):
		logger.debug("starting with execute")
		if self._is_executing:
			logger.debug("nested execute call")
			# this situation may happen since in this methods, via a callback (to update a progressbar) we enter
			# Qt's eventloop, which may execute code that will call execute again
			# as long as that code is using delay tasks (i.e. promises) we can simple return here, since after
			# the execute is almost finished, any new tasks added to the task_queue will get executing
			return
		# u 'column' is uniquely identified by a tuple of (dataset, expression)
		self._is_executing = True
		try:
			t0 = time.time()
			task_queue_all = list(self.task_queue)
			if not task_queue_all:
				logger.info("only had cancelled tasks")
			logger.info("clearing queue")
			#self.task_queue = [] # Ok, this was stupid.. in the meantime there may have been new tasks, instead, remove the ones we copied
			for task in task_queue_all:
				logger.info("remove from queue: %r", task)
				self.task_queue.remove(task)
			logger.info("left in queue: %r", self.task_queue)
			task_queue_all = [task for task in task_queue_all if not task.cancelled]
			logger.debug("executing queue: %r" % (task_queue_all))

			#for task in self.task_queue:
			#$	print task, task.expressions_all
			datasets = set(task.dataset for task in task_queue_all)
			cancelled = [False]
			def cancel():
				logger.debug("cancelling")
				self.signal_cancel.emit()
				cancelled[0] = True
			try:
				# process tasks per dataset
				self.signal_begin.emit()
				for dataset in datasets:
					task_queue = [task for task in task_queue_all if task.dataset == dataset]
					expressions = list(set(expression for task in task_queue for expression in task.expressions_all))

					for task in task_queue:
						task._results = []
						task.signal_progress.emit(0)
					self.server.execute_queue(task_queue)
					self._is_executing = False
			except:
				# on any error we flush the task queue
				self.signal_cancel.emit()
				logger.exception("error in task, flush task queue")
				raise
			logger.debug("executing took %r seconds" % (time.time() - t0))
			# while processing the self.task_queue, new elements will be added to it, so copy it
			logger.debug("cancelled: %r", cancelled)
			if cancelled[0]:
				logger.debug("execution aborted")
				task_queue = task_queue_all
				for task in task_queue:
					#task._result = task.reduce(task._results)
					#task.reject(UserAbort("cancelled"))
					# remove references
					task._result = None
					task._results = None
			else:
				task_queue = task_queue_all
				for task in task_queue:
					logger.debug("fulfill task: %r", task)
					if not task.cancelled:
						task._result = task.reduce(task._results)
						task.fulfill(task._result)
						# remove references
					task._result = None
					task._results = None
				self.signal_end.emit()
				# if new tasks were added as a result of this, execute them immediately
				# TODO: we may want to include infinite recursion protection
				self._is_executing = False
				if len(self.task_queue) > 0:
					logger.debug("task queue not empty.. start over!")
					self.execute()
		finally:
			self._is_executing = False

if __name__ == "__main__":
	import vaex
	import sys
	vaex.set_log_level_debug()
	server = vaex.server(sys.argv[1], port=int(sys.argv[2]))
	datasets = server.datasets()
	print(datasets)
	dataset = datasets[0]
	dataset = vaex.example()
	print(dataset("x").minmax())
	dataset.select("x < 0")
	print(dataset.selected_length(), len(dataset))
	print(dataset("x").selected().is_masked)
	print(dataset("x").selected().minmax())



# -*- coding: utf-8 -*-

import logging as logging
logger = logging.getLogger("vaex.kld")
import numpy as np
import math
import functools
import vaex.vaexfast
import vaex.utils
from vaex.utils import Timer

def mutual_information(data):
	Q = vaex.utils.disjoined(data)
	P = data

	P = P / P.sum()
	Q = Q / Q.sum()
	mask = (P > 0) & (Q>0)
	information = np.sum(P[mask] * np.log(P[mask]/Q[mask]))# * np.sum(dx)
	return information

def kl_divergence(P, Q, axis=None):
    P = P / P.sum(axis=axis)
    Q = Q / Q.sum(axis=axis)
    #mask = (P > 0) & (Q>0)
    #information = np.sum(P[mask] * np.log(P[mask]/Q[mask]), axis=axis)# * np.sum(dx)
    #mask = (P > 0) & (Q>0)
    information = np.sum(P * np.log(P/Q), axis=axis)# * np.sum(dx)
    return information

class KlDivergenceShuffle(object):
	def __init__(self, dataset, pairs, gridsize=128):
		self.dataset = dataset
		self.pairs = pairs
		self.dimension = len(self.pairs[0])
		self.logger = logger.getLogger('kld')
		self.gridsize = gridsize
		logger.debug("dimension: %d, pairs: %s" % (self.dimension, self.pairs))
		
		
	def get_jobs(self):
		def job(pair):
			pass
		
		
def to_disjoined(counts):
	shape = counts.shape
	assert len(counts.shape) == 2
	counts_0 = counts.sum(axis=1).reshape((shape[0], 1))
	counts_1 = counts.sum(axis=0).reshape((1, shape[1]))
	counts_disjoined = counts_0 * counts_1
	return counts_disjoined
		
def kld_shuffled(columns, Ngrid=128, datamins=None, datamaxes=None, offset=1):
	if datamins is None:
		datamins = np.array([np.nanmin(column) for column in columns])
	if datamaxes is None:
		datamaxes = np.array([np.nanmax(column) for column in columns])
	dim = len(columns)
	counts = np.zeros((Ngrid, ) * dim, dtype=np.float64)
	counts_shuffled = np.zeros((Ngrid, ) * dim, dtype=np.float64)
	D_kl = -1
	if len(columns) == 2:
		x, y = columns
		#print x
		#print y
		print((x, y, counts, counts_shuffled, datamins[0], datamaxes[0], datamins[1], datamaxes[1], offset))
		try:
			vaex.histogram.hist2d_and_shuffled(x, y, counts, counts_shuffled, datamins[0], datamaxes[0], datamins[1], datamaxes[1], offset)
		except:
			args = [x, y, counts, counts_shuffled, datamins[0], datamaxes[0], datamins[1], datamaxes[1], offset]
			sig = [numba.dispatcher.typeof_pyval(a) for a in args]
			print(sig)
			raise
			
		print(("counts", sum(counts)))
		deltax = [float(datamaxes[i] - datamins[i]) for i in range(dim)]
		dx = np.array([deltax[d]/counts.shape[d] for d in range(dim)])
		density = counts/np.sum(counts)# * np.sum(dx)
		density_shuffled = counts_shuffled / np.sum(counts_shuffled)# * np.sum(dx)
		mask = (density_shuffled > 0) & (density>0)
		#print density
		D_kl = np.sum(density[mask] * np.log(density[mask]/density_shuffled[mask]))# * np.sum(dx)
		#if D_kl < 0:
		#	import pdb
		#	pdb.set_trace()
	return D_kl
				
		
def kld_shuffled_grouped(dataset, range_map, pairs, feedback=None, size_grid=32, use_mask=True, bytes_max=int(1024**3/2)):
	dimension = len(pairs[0])
	bytes_per_grid = size_grid ** dimension * 8 # 8==sizeof (double)
	grids_per_iteration = min(128, int(bytes_max/bytes_per_grid))
	iterations = int(math.ceil(len(pairs) * 1. / grids_per_iteration))
	jobsManager = vaex.dataset.JobsManager()
	ranges = [None] * len(pairs)
	D_kls = []


	class Wrapper(object):
		pass
	wrapper = Wrapper()
	wrapper.N_done = 0

	counts = np.zeros((grids_per_iteration,) + (size_grid,) * dimension, dtype=np.float64)
	#counts_shuffled = np.zeros((grids_per_iteration,) + (size_grid,) * dimension, dtype=np.float64)

	N_total = len(pairs) * len(dataset) * dimension + len(pairs) *counts.size

	logger.debug("{iterations} iterations with {grids_per_iteration} grids per iteration".format(**locals()))
	for part in range(iterations):
		if part > 0: # next iterations reset the counts
			counts.reshape(-1)[:] = 0
			#counts_shuffled.reshape(-1)[:] = 0
		i1, i2 = part * grids_per_iteration, (part+1)*grids_per_iteration
		if i2 > len(pairs):
			i2 = len(pairs)
		n_grids = i2 - i1
		logger.debug("part {part} of {iterations}, from {i1} to {i2}".format(**locals()))
		def grid(info, *blocks, **kwargs):
			index = kwargs["index"]
			if use_mask and dataset.mask is not None:
				mask = dataset.mask[info.i1:info.i2]
				blocks = [block[mask] for block in blocks]
			else:
				mask = None
			ranges = []
			minima = []
			maxima = []
			for dim in range(dimension):
				ranges += list(range_map[pairs[index][dim]])
				mi, ma = range_map[pairs[index][dim]]
				minima.append(mi)
				maxima.append(ma)
			if len(blocks) == 2:
				print(("mask", mask))
				vaex.vaexfast.histogram2d(blocks[0], blocks[1], None, counts[index], *(ranges + [0, 0]))
				#vaex.vaexfast.histogram2d(blocks[0], blocks[1], None, counts_shuffled[index], *(ranges + [1, 0]))
			if len(blocks) == 3:
				vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], None, counts[index], *(ranges + [0,0,0]))
				#vaex.vaexfast.histogramNd(list(blocks), None, counts[index], minima, maxima)
				#for i in range(5):
				#	vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], None, counts_shuffled[index], *(ranges + [2+i,1+i,0]))
			if len(blocks) > 3:
				vaex.vaexfast.histogramNd(list(blocks), None, counts[index], minima, maxima)
				#for i in range(5):
				#	vaex.vaexfast.histogram3d(blocks[0], blocks[1], blocks[2], None, counts_shuffled[index], *(ranges + [2+i,1+i,0]))
			if feedback:
				wrapper.N_done += len(blocks[0]) * dimension
				if feedback:
					cancel = feedback(wrapper.N_done*100./N_total)
					if cancel:
						raise Exception("cancelled")
				
		for index, pair in zip(list(range(i1, i2)), pairs[i1:i2]):
			logger.debug("add job %r %r" % (index, pair))
			jobsManager.addJob(0, functools.partial(grid, index=index-i1), dataset, *pair) 
		jobsManager.execute()

	
		with Timer("D_kl"):
			for i in range(n_grids):
				if 0:
					deltax = [float(range_map[pairs[i][d]][1] - range_map[pairs[i][d]][0]) for d in range(dimension)]
					dx = np.array([deltax[d]/counts[i].shape[d] for d in range(dimension)])
					density = counts[i]/np.sum(counts[i])# * np.sum(dx)
					counts_shuffled = to_disjoined(counts[i])
					density_shuffled = counts_shuffled / np.sum(counts_shuffled)# * np.sum(dx)
					mask = (density_shuffled > 0) & (density>0)
					print(("mask sum", np.sum(mask)))
					print(("mask sum", np.sum((counts_shuffled > 0) & (counts[i]>0))))
					#print density
					D_kl = np.sum(density[mask] * np.log(density[mask]/density_shuffled[mask]))# * np.sum(dx)
				else:
					D_kl = mutual_information(counts[i])
				D_kls.append(D_kl)
				if feedback:
					wrapper.N_done += counts.size
					cancel = feedback(wrapper.N_done*100./N_total)
					if cancel:
						return None
						#raise Exception, "cancelled"
	return D_kls
			
		
		
	
if __name__ == "__main__":
	import vaex.dataset
	import vaex.files
	from optparse import OptionParser
	parser = OptionParser() #usage="")

	#parser.add_option("-n", "--name",
	#                 help="dataset name [default=%default]", default="data", type=str)
	#parser.add_option("-o", "--output",
	#                 help="dataset output filename [by default the suffix of input filename will be replaced by hdf5]", default=None, type=str)
	(options, args) = parser.parse_args()
	if len(args) > 0:
		filename = args[0]
	else:
		filename = "gaussian4d-1e7.hdf5"
	
	path = vaex.files.get_datafile(filename)
	dataset = vaex.dataset.Hdf5MemoryMapped(path)
	#for column_name in dataset.column_names:
	#	print column_name
	
	subspace = dataset.subspace(1)[:3] * dataset.subspace(1)[:3]
	#subspace = Subspace(dataset, [("x",), ("y",), ("z",)]


import vaex as vx
import sys
import timeit
import numpy as np
import argparse

subspace, dataset, limits = None, None, None

def main(argv):
	global subspace, dataset, limits
	parser = argparse.ArgumentParser(argv[0])
	parser.add_argument("-N", help="run each batch N times (default: %(default)s)", type=int, default=5)
	parser.add_argument("-f", "--fraction", help="fraction of dataset to use (default: %(default)s)", default=1., type=float)
	parser.add_argument("-b", "--buffer-size", default=vx.execution.buffer_size_default, help="buffer size per thread (default: %(default)s)", type=int)
	parser.add_argument("-r", "--repeat", default=5, help="repeat benchmark X times (default: %(default)s)", type=int)
	parser.add_argument("-c", "--cpu-count", default=vx.multithreading.thread_count_default, help="thread count/cpu count (default: %(default)s)", type=int)
	parser.add_argument("filename", help="filename of dataset to use")
	parser.add_argument("expressions", help="list of expressions to export (or all when empty)", nargs="*")

	args = parser.parse_args(argv[1:])
	vx.execution.buffer_size_default = args.buffer_size
	vx.multithreading.thread_count_default = args.cpu_count

	progressbar = False
	fn = args.filename
	#print(("opening", fn))
	dataset = vx.open(fn)
	dataset.set_active_fraction(args.fraction)
	#dataset = vx.open_many(fn)

	expressions = args.expressions
	#print "subspace", expressions
	subspace = dataset(*expressions)
	byte_size = len(dataset) * len(expressions) * 8
	#sums = subspace.sum()

	limits = subspace.minmax()
	print(limits)

	N = args.N
	print("benchmarking minmax")
	expr = "subspace.minmax()"
	times = timeit.repeat(expr, setup="from vaex.benchmark import subspace, dataset, np", repeat=args.repeat, number=N)
	print("minimum time", min(times)/N)
	bandwidth = [byte_size/1024.**3/(time/N) for time in times]
	print("%f GiB/s" % max(bandwidth))

	speed = [len(dataset)/(time/N)/1e9 for time in times]
	print("%f billion rows/s " % max(speed))

	print()
	print("benchmarking histogram")
	expr = "dataset.count(binby=['{}', '{}'])".format(*expressions)
	times = timeit.repeat(expr, setup="from vaex.benchmark import subspace, dataset, np, limits", repeat=args.repeat, number=N)
	print("minimum time", min(times)/N)
	bandwidth = [byte_size/1024.**3/(time/N) for time in times]
	print("%f GiB/s" % max(bandwidth))

	speed = [len(dataset)/(time/N)/1e9 for time in times]
	print("%f billion rows/s " % max(speed))
	print()

	#expr = "np.nansum(dataset.columns['x'])"
	print("sum=", subspace.sum())
	print("benchmarking sum")
	expr = "subspace.sum()"
	times = timeit.repeat(expr, setup="from vaex.benchmark import subspace, dataset, np", repeat=args.repeat, number=N)
	print("minimum time", min(times)/N)
	bandwidth = [byte_size/1024.**3/(time/N) for time in times]
	print("%f GiB/s" % max(bandwidth))




from __future__ import absolute_import

__author__ = 'maartenbreddels'
import numpy as np
import vaex.image
import vaex.dataset
import logging
import vaex as vx
import vaex.delayed
from .common import Job
import os
from IPython.display import HTML, display_html, display_javascript, display
import bqplot.marks
import bqplot as bq
import bqplot.interacts
import traitlets
import ipywidgets as widgets
from vaex.jupyter import debounced
import vaex.grids
import time


logger = logging.getLogger("vaex.ext.bqplot")
base_path = os.path.dirname(__file__)

@bqplot.marks.register_mark('vaex.ext.bqplot.Image')
class Image(bqplot.marks.Mark):
    src = bqplot.marks.Unicode().tag(sync=True)
    x = bqplot.marks.Float().tag(sync=True)
    y = bqplot.marks.Float().tag(sync=True)
    view_count = traitlets.CInt(0).tag(sync=True)
    width = bqplot.marks.Float().tag(sync=True)
    height = bqplot.marks.Float().tag(sync=True)
    preserve_aspect_ratio = bqplot.marks.Unicode('').tag(sync=True)
    _model_module = bqplot.marks.Unicode('vaex.ext.bqplot').tag(sync=True)
    _view_module = bqplot.marks.Unicode('vaex.ext.bqplot').tag(sync=True)

    _view_name = bqplot.marks.Unicode('Image').tag(sync=True)
    _model_name = bqplot.marks.Unicode('ImageModel').tag(sync=True)
    scales_metadata = bqplot.marks.Dict({
        'x': {'orientation': 'horizontal', 'dimension': 'x'},
        'y': {'orientation': 'vertical', 'dimension': 'y'},
    }).tag(sync=True)

    def __init__(self, **kwargs):
        self._drag_end_handlers = bqplot.marks.CallbackDispatcher()
        super(Image, self).__init__(**kwargs)


import warnings

patched = False
def patch(force=False):
    # return
    global patched
    if not patched or force:
        display_javascript(open(os.path.join(base_path, "bqplot_ext.js")).read(), raw=True)
    patched = True

# if (bqplot.__version__ == (0, 6, 1)) or (bqplot.__version__ == "0.6.1"):
# else:
#	warnings.warn("This version (%s) of bqplot is not supppored" % bqplot.__version__)


if 0:
    class _BqplotHistogram(Plot2d):
        def __init__(self, subspace, color, size, limits):
            self.color = color
            super(BqplotHistogram, self).__init__(subspace, size, limits)

        def create(self, data):
            size = data.shape[0]
            assert len(data.shape) == 1
            xmin, xmax = self.limits[0]
            dx = (xmax - xmin) / size
            x = np.linspace(xmin, xmax - dx, size) + dx / 2
            # print xmin, xmax, x

            self.scale_x = bq.LinearScale(min=xmin + dx / 2, max=xmax - dx / 2)
            self.scale_y = bq.LinearScale()

            self.axis_x = bq.Axis(label='X', scale=self.scale_x)
            self.axis_y = bq.Axis(label='Y', scale=self.scale_y, orientation='vertical')
            self.bars = bq.Bars(x=x,
                                y=data, scales={'x': self.scale_x, 'y': self.scale_y}, colors=[self.color])

            self.fig = bq.Figure(axes=[self.axis_x, self.axis_y], marks=[self.bars], padding_x=0)

        def update(self, data):
            self.bars.y = data


    def BqplotHistogram2d(Bqplot):
        def __init__(self, subspace, color, size, limits):
            self.color = color
            super(BqplotHistogram, self).__init__(subspace, size, limits)

        def create(self, data):
            pass




import logging

logger = logging.getLogger("vaex.settings")

class Files(object):
	def __init__(self, open, recent):
		self.open = open
		self.recent = recent

import yaml
from yaml import Loader, Dumper

class Settings(object):
	def __init__(self, filename):
		self.filename = filename
		if not os.path.exists(filename):
			with open(filename, "w"):
				pass
		with open(self.filename) as f:
			self.settings = vaex.utils.yaml_load(f) #yaml.load(f, Loader=Loader)
		if self.settings is None:
			self.settings = {}
		#logger.debug("settings: %r", self.settings)

	def auto_store_dict(self, key):
		# TODO: no nested keys supported yet
		if key not in self.settings:
			self.settings[key] = {}
		return AutoStoreDict(self, self.settings[key])

	def store(self, key, value):
		parts = key.split(".")
		obj = self.settings
		for part in parts[:-1]:
			#print part,
			if part not in obj:
				obj[part] = {}
			obj = obj[part]
		obj[parts[-1]] = value
		#print self.settings
		self.dump()

	def dump(self):
		with open(self.filename, "w") as f:
			#yaml.dump(self.settings, f)
			vaex.utils.yaml_dump(f, self.settings)

	def get(self, key, default=None):
		logger.debug("get %r", key)
		parts = key.split(".")
		obj = self.settings
		for part in parts:
			if part not in obj:
				logger.debug("return %r (default)", default)
				return default
			obj = obj[part]
		logger.debug("return %r", obj)
		return obj

settings = {}

import vaex.utils
import os

import collections


class AutoStoreDict(collections.MutableMapping):
	def __init__(self, settings, store):
		self.store = store
		self.settings = settings

	def __getitem__(self, key):
		return self.store[self.__keytransform__(key)]

	def __setitem__(self, key, value):
		self.store[self.__keytransform__(key)] = value
		self.settings.dump()

	def __delitem__(self, key):
		del self.store[self.__keytransform__(key)]
		self.settings.dump()

	def __iter__(self):
		return iter(self.store)

	def __len__(self):
		return len(self.store)

	def __keytransform__(self, key):
		return key


main = Settings(os.path.join(vaex.utils.get_private_dir(), "main.yml"))
webclient = Settings(os.path.join(vaex.utils.get_private_dir(), "webclient.yml"))
webserver = Settings(os.path.join(vaex.utils.get_private_dir(), "webserver.yml"))
cluster = Settings(os.path.join(vaex.utils.get_private_dir(), "cluster.yml"))

#yaml.load()

if __name__ == "__main__":
	webclient.store("bla.la.l", 1)


# -*- coding: utf-8 -*-
#from sampy import *
#from SocketServer import ThreadingMixIn
#try:
#	import sampy
#except ImportError:
#import astropy.samp as sampy
import logging
import threading
import time
import astropy.samp

logger = logging.getLogger("vaex.samp")


class Samp(object):
	def __init__(self, daemon=True, name=None):
		self.client = astropy.samp.SAMPIntegratedClient(metadata = {"samp.name":"Gavi client" if name is None else name,
										"samp.description.text": "Gavi client" if name is None else name,
										"gavi.samp.version":"0.01"}, callable=True)


		# sampy doesn't make this thread Daeamon, so the python process never stops on the cmd line
		# this fixes that
		#def _myrun_client():
		#	if self.client.client._callable:
		#		self.client.client._thread = threading.Thread(target = self.client.client._serve_forever)
		#		self.client.client._thread.setDaemon(True)
		#		self.client.client._thread.start()
		#if daemon:
		#	self.client.client._run_client = _myrun_client
		connected = False
		try:
			self.client.connect()
			connected = True
		except astropy.samp.SAMPHubError as e:
			#print "error connecting to hub", e
			pass

		if connected:
			#self.client.client._thread.setDaemon(False)
			logger.info("connected to SAMP hub")
			logger.info("binding events")
			self.client.bind_receive_call			("table.load.votable", self._onTableLoadVotable)
			self.client.bind_receive_notification	("table.load.votable", self._onTableLoadVotable)
			#self.client.bindReceiveNotification	("table.highlight.row", self._onSampNotification)
			#self.client.bindReceiveMessage("table.load.votable", self._onSampCall)
			#self.client.bindReceiveResponse("table.load.votable", self._onSampCall)
			
			#self.client.bindReceiveCall("samp.*", self._onSampCall)
			#self.client.bindReceiveNotification("samp.*", self._onSampNotification)
			
			#self.client.bindReceiveCall("table.*", self._onSampCall)
			#self.client.bindReceiveNotification("table.*", self._onSampNotification)
			#self.client.bindReceiveMessage("table.*", self._onSampCall)
			#self.client.bindReceiveResponse("table.*", self._onSampCall)
			
			#self.client.bindReceiveMessage("table.votable.*", self._onSampCall)
			#self.client.bindReceiveResponse("table.votable.*", self._onSampCall)
			
	#def connect(self):
	#	self.client.connect()
		self.tableLoadCallbacks = []
	
	def _onTableLoadVotable(self, private_key, sender_id, msg_id, mtype, params, extra):
		print(("Msg:", repr(private_key), repr(sender_id), repr(msg_id), repr(mtype), repr(params), repr(extra)))
		try:
			url = params["url"]
			table_id = params["table-id"]
			name = params["name"]
			for callback in self.tableLoadCallbacks:
				callback(url, table_id, name)
		except:
			logger.exception("event handler failed")
		
		if msg_id != None: # if SAMP call, send a reply
			self.client.ereply(msg_id, sampy.SAMP_STATUS_OK, result = {"txt": "loaded"})
		
	def _onSampNotification(self, private_key, sender_id, mtype, params, extra):
		print(("Notification:", repr(private_key), repr(sender_id), repr(mtype), repr(params), repr(extra)))
		
	def _onSampCall(self, private_key, sender_id, msg_id, mtype, params, extra):
		print("----")
		try:
			print(("Call:", repr(private_key), repr(sender_id), repr(msg_id), repr(mtype), repr(params), repr(extra)))
			self.client.ereply(msg_id, sampy.SAMP_STATUS_OK, result = {"txt": "printed"})
		except:
			print("errrrrrrororrrr hans!")



# similar to http://astrofrog-debug.readthedocs.io/en/latest/vo/samp/example_table_image.html
class SampSingle(object):
	def __init__(self, name="vaex - single table load"):
		self.done = False
		self.client = astropy.samp.SAMPIntegratedClient(name=name)
		self.client.connect()

		def call(private_key, sender_id, msg_id, mtype, params, extra):
			self.params = params
			self.client.reply(msg_id, {"samp.status": "samp.ok", "samp.result": {}})
			self.done = True
		def notify(private_key, sender_id, mtype, params, extra):
			self.params = params
			self.done = True

		self.client.bind_receive_call("table.load.votable", call)
		self.client.bind_receive_notification("table.load.votable", notify)



	def wait_for_table(self):
		logging.debug("waiting for samp msg: table.load.votable")
		while not self.done:
			time.sleep(0.1)
		logging.debug("got samp msg: table.load.votable with params: %s", self.params)
		self.client.disconnect()
		return self.params["url"]


import getpass
import requests
from requests.auth import HTTPBasicAuth
import astropy.io.votable

try:
	input = raw_input
except:
	pass
# py2/3 comp
try:
    from StringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO

def ask_cmd_line(username, password):
	if not username:
		username = input("Username:")
	if not password:
		password = getpass.getpass("Password:")
	return username, password

def fetch_votable(url, username=None, password=None, ask=ask_cmd_line):
	done = False
	auth = None
	while not done:
		if username and password:
			auth = HTTPBasicAuth(username, password)
		req = requests.get(url, auth=auth)
		if req.status_code == 401:
			result = ask(username, password)
			if result:
				username, password = result
				auth = HTTPBasicAuth(username, password)
			else:
				return
		else:
			done = True
	if req.status_code == 200:
		f = BytesIO()
		f.write(req.content)
		return astropy.io.votable.parse_single_table(f)
	else:
		return req

def single_table(username=None, password=None):
	#from vaex.samp import SampSingle
	#from astropy.table import Table
	samp = SampSingle()
	url = samp.wait_for_table()
	import os
	import tempfile
	import requests

	done = False
	table = fetch_votable(url, username, password)
	return table
	# while not done:
	# r = requests.get()
	# tempdir = tempfile.mkdtemp()
	# path = os.path.join(tempdir, "out.vot.gz")
	# os.system("wget -c -O %s %s" % (path, url))
	# parts = os.path.split(url)
	# print("parts", parts)
	# path2 = os.path.join(tempdir, "out.vot")
	# os.system("gunzip %s > /dev/null; mv %s %s  > /dev/null" % (path, path, path2))
	# print(path, path2)
	# table = Table.read(path2, format="votable")
	# return from_astropy_table(table)



# -*- coding: utf-8 -*-
import logging
import logging
import collections

logger = logging.getLogger("expr")
logger.setLevel(logging.ERROR)

class Base(object):
	def __neg__(self):
		return Neg(self)
	def __add__(self, rhs):
		return Add(self, rhs)
	def __radd__(self, lhs):
		return Add(lhs, self)
	def __sub__(self, rhs):
		return Sub(self, rhs)
	def __rsub__(self, lhs):
		return Sub(lhs, self)
	def __mul__(self, rhs):
		return Mul(self, rhs)
	def __rmul__(self, lhs):
		return Mul(lhs, self)
	def __div__(self, rhs):
		return Div(self, rhs)
	def __pow__(self, rhs):
		return Pow(self, rhs)

	def walk(self, f):
		return f(self)

class Sqrt(Base):
	def __init__(self, value):
		self.value = check_const(value)

	def __repr__(self):
		return "sqrt(" + repr(self.value) + ")"
class Const(Base):
	def __init__(self, value):
		self.value = value

	def __repr__(self):
		return repr(self.value)

	def c_code(self):
		return repr(self.value)

def check_const(value):
	if isinstance(value, (int, float)):
		return Const(value)
	else:
		return value

class Neg(Base):
	def __init__(self, obj):
		self.obj = check_const(obj)

	def __repr__(self):
		return "(-" +repr(self.obj) + ")"

	def walk(self, f):
		#logger.debug("walk neg")
		return f(Neg(self.obj.walk(f)))

class Add(Base):
	def __init__(self, lhs, rhs):
		self.lhs = check_const(lhs)
		self.rhs = check_const(rhs)

	def __repr__(self):
		return "(" +repr(self.lhs) +" + " +repr(self.rhs) +")"

	def walk(self, f):
		return f(Add(self.lhs.walk(f), self.rhs.walk(f)))

	def c_code(self):
		return "(" +self.lhs.c_code() +" + " +self.rhs.c_code() +")"

class Sub(Base):
	def __init__(self, lhs, rhs):
		self.lhs = check_const(lhs)
		self.rhs = check_const(rhs)

	def __repr__(self):
		return "(" +repr(self.lhs) +" - " +repr(self.rhs) +")"

	def walk(self, f):
		return f(Sub(self.lhs.walk(f), self.rhs.walk(f)))

class Div(Base):
	def __init__(self, lhs, rhs):
		self.lhs = check_const(lhs)
		self.rhs = check_const(rhs)

	def __repr__(self):
		return "(" +repr(self.lhs) +" / " +repr(self.rhs) +")"

	def walk(self, f):
		return f(Div(self.lhs.walk(f), self.rhs.walk(f)))

class Mul(Base):
	def __init__(self, lhs, rhs):
		self.lhs = check_const(lhs)
		self.rhs = check_const(rhs)

	def __repr__(self):
		return "(" +repr(self.lhs) +" * " +repr(self.rhs) +")"

	def c_code(self):
		return "(" +self.lhs.c_code() +" * " +self.rhs.c_code() +")"

	def walk(self, f):
		return f(Mul(self.lhs.walk(f), self.rhs.walk(f)))

class Pow(Base):
	def __init__(self, lhs, rhs):
		self.lhs = check_const(lhs)
		self.rhs = check_const(rhs)

	def __repr__(self):
		return "(" +repr(self.lhs) +" ** " +repr(self.rhs) +")"

	def walk(self, f):
		return f(Pow(self.lhs.walk(f), self.rhs.walk(f)))

class Function(Base):
	def __init__(self, var, args):
		self.var = var
		self.args = args
		#print args

	def __repr__(self):
		return self.var.name +"(" +", ".join([repr(k) for k in self.args]) + ")"

	def walk(self, f):
		args = [arg.walk(f) for arg in self.args]
		return f(Function(self.var.walk(f), args))


class Slice(Base):
	def __init__(self, var, args):
		self.var = var
		self.args = args
		try:
			len(self.args)
		except:
			self.args = (self.args,)

	def tovar(self):
		parts = [self.var.name]
		for arg in self.args:
			if isinstance(arg, (int, float)):
				parts.append(repr(arg).replace("-", "_min_"))
			elif isinstance(arg, Var):
				parts.append(arg.name)
			elif arg is None:
				parts.append("None")
			else:
				s = ""
				s += repr(arg.start)
				s += "_"
				s += repr(arg.stop)
				s += "_"
				s += repr(arg.step)
				parts.append(s)
		return "__".join(parts)


	def __repr__(self):
		parts = []
		#logger.debug("args: %r" % self.args)
		for arg in self.args:
			if isinstance(arg, (int, float)):
				parts.append(repr(arg))
			elif isinstance(arg, Var):
				parts.append(arg.name)
			elif arg is None:
				parts.append("")
			else:
				s = ""
				if arg.start is not None:
					s += repr(arg.start)
				s += ":"
				if arg.stop is not None:
					s += repr(arg.stop)
				s += ":"
				if arg.step is not None:
					s += repr(arg.step)
				if s == "::":
					s = ":"
				parts.append(s)
		return self.var.name + "[" +",".join(parts) + "]"

class Var(Base):
	def __init__(self, name):
		self.name = name

	def __getitem__(self, args):
		logger.debug("[%s] Var.__getitem__(%r)" % (self.name, args,))
		return Slice(self, args)

	def __call__(self, *args):
		logger.debug("[%s] Var.__call__(%r)" % (self.name, args,))
		return Function(self, args)

	def __str__(self):
		return "Var(%r)" % self.name
	def __repr__(self):
		return self.name

	def c_code(self):
		return self.name

	def walk(self, f):
		return f(self)

class Scope(object):
	def __init__(self, name):
		self.name = name
		self.resolved = collections.OrderedDict()

	def write(self, f):
		for block_index, (key, value) in enumerate(self.resolved.items()):
			# this assumes the var is part of a 'block'
			f.write("double %s = blocks[%i][index];\n" % (key, block_index) )

	def __getitem__(self, name):
		logger.debug("[%s] Scope.__getitem__(%r)" % (self.name, name,))
		if name not in self.resolved:
			self.resolved[name] = Var(name)
		else:
			logger.debug("[%s] Scope.__getitem__(%r) (resolved again)" % (self.name, name,))
		return self.resolved[name]

		#return Scope(name)

def distance(args):
	if len(args) == 1:
		result = args[0]
	else:
		result = Add(args[0]**2, args[1]**2)
		for arg in args[2:]:
			result = Add(result, arg**2)
		result = Sqrt(result)
	return result
macros = {"l2": distance}
def translate(expression, replacements={}):
	scope = Scope(None)
	newvars = {}
	used_vars = set()
	result = eval(expression, {}, scope)
	#for key, value in replacements:
	translated_replacements = {}
	for i, (key, expression) in enumerate(replacements.items()):
		current_replacements = collections.OrderedDict(list(replacements.items())[:i])
		#print "*" * 70
		translated_replacements[key] = translate(expression, current_replacements)[0]
		#print "TRANSLATE:", expression, translated_replacements[key], current_replacements

	def slice_to_var(slice):
		newname = slice.tovar()
		logger.debug("replace %s by %s" % (slice.var.name, newname))
		newvars[newname] = slice
		return Var(newname)

	def walker(expr):
		#print ">>>>", expr, type(expr)
		if isinstance(expr, Slice):
			return slice_to_var(expr)
		elif isinstance(expr, Function):
			#logger.debug("expr: " + expr.var.name)
			if expr.var.name in macros:
				return macros[expr.var.name](expr.args)
			else:
				return expr
		elif isinstance(expr, Var):
			#logger.debug("var: " + expr.name)
			used_vars.add(expr.name)
			if expr.name in translated_replacements:
				return translated_replacements[expr.name]
			else:
				return expr
		else:
			return expr
	#print "walking.."
	newresult = result.walk(walker)
	return newresult, newvars #, used_vars

class MathExpression(object):
	def __init__(self, expression_string, macros):
		self.expression_string = expression_string
		self.scope = Scope()
		self.macros = macros
		for macro in self.macros:
			scope.add_macro(macro)
		self.expression = eval(self.expression_string)

	def evaluate(self, values):
		return eval(self.expression, math.__dict__, values)

	def evaluate_numpy(self, values):
		return eval(self.expression, np.__dict__, values)

	def variables(self):
		eval(self.expression, np.__dict__, scope)

import ast
import _ast
import string
import numpy as np
import math
import sys

valid_binary_operators = [_ast.Add, _ast.Sub, _ast.Mult, _ast.Pow, _ast.Div, _ast.FloorDiv, _ast.BitAnd, _ast.BitOr, _ast.BitXor, _ast.Mod]
valid_compare_operators = [_ast.Lt, _ast.LtE, _ast.Gt, _ast.GtE, _ast.Eq, _ast.NotEq]
valid_unary_operators = [_ast.USub, _ast.UAdd]
valid_id_characters = string.ascii_letters + string.digits + "_"
valid_functions = "sin cos".split()

def math_parse(expression, macros=[]):
	# TODO: validate macros?
	node = ast.parse(expression)
	if len(node.body) != 1:
		raise ValueError("expected one expression, got %r" % len(node.body))
	expr = node.body[0]
	if not isinstance(expr, _ast.Expr):
		raise ValueError("expected an expression got a %r" % type(node.body))

	validate_expression(expr.value)
	return MathExpression(expression, macros)
import six
last_func = None
def validate_expression(expr, variable_set, function_set=[], names=None):
	global last_func
	names = names if names is not None else []
	if isinstance(expr, six.string_types):
		node = ast.parse(expr)
		if len(node.body) != 1:
			raise ValueError("expected one expression, got %r" % len(node.body))
		first_expr = node.body[0]
		if not isinstance(first_expr, _ast.Expr):
			raise ValueError("expected an expression got a %r" % type(node.body))
		validate_expression(first_expr.value, variable_set, function_set, names)
	elif isinstance(expr, _ast.BinOp):
		if expr.op.__class__ in valid_binary_operators:
			validate_expression(expr.right, variable_set, function_set, names)
			validate_expression(expr.left, variable_set, function_set, names)
		else:
			raise ValueError("Binary operator not allowed: %r" % expr.op)
	elif isinstance(expr, _ast.UnaryOp):
		if expr.op.__class__ in valid_unary_operators:
			validate_expression(expr.operand, variable_set, function_set, names)
		else:
			raise ValueError("Unary operator not allowed: %r" % expr.op)
	elif isinstance(expr, _ast.Name):
		validate_id(expr.id)
		if expr.id not in variable_set:
			raise NameError("variable %r is not defined (available are: %s)" % (expr.id, ", ".join(list(variable_set))))
		names.append(expr.id)
	elif isinstance(expr, _ast.Num):
		pass # numbers are fine
	elif isinstance(expr, _ast.Call):
		validate_func(expr.func, function_set)
		last_func = expr
		for arg in expr.args:
			validate_expression(arg, variable_set, function_set, names)
		for arg in expr.keywords:
			validate_expression(arg, variable_set, function_set, names)
	elif isinstance(expr, _ast.Compare):
		validate_expression(expr.left, variable_set, function_set, names)
		for op in expr.ops:
			if op.__class__ not in valid_compare_operators:
				raise ValueError("Compare operator not allowed: %r" % op)
		for comparator in expr.comparators:
			validate_expression(comparator, variable_set, function_set, names)
	elif isinstance(expr, _ast.keyword):
		validate_expression(expr.value, variable_set, function_set, names)
	elif isinstance(expr, _ast.Subscript):
		validate_expression(expr.value, variable_set, function_set, names)
		if isinstance(expr.slice.value, _ast.Num):
			pass # numbers are fine
		else:
			raise ValueError("Only subscript/slices with numbers allowed, not: %r" % expr.slice.value)
	else:
		last_func = expr
		raise ValueError("Unknown expression type: %r" % type(expr))

def validate_func(name, function_set):
	if name.id not in function_set:
		raise NameError("function %r is not defined" % name.id)
	#if name.id not in valid_functions:
	#	raise ValueError, "invalid or unknown function %r" % name.id


def validate_id(id):
	for char in id:
		if char not in valid_id_characters:
			raise ValueError("invalid character %r in id %r" % (char, id))

expressions = [expr for expr in """
a * b
a + b
sin(a)
a**b - sin(a**2, b--a)
""".split("\n") if expr]

if __name__ == "__main__":
	values = dict(a=1, b=2, pi=math.pi)
	for expression in expressions:
		ex = math_parse(expression)
		print((expression, "=", ex.evaluate(values)))
	values = dict(a=np.array([1,2]), b=np.array([2,3]), pi=math.pi)
	for expression in expressions:
		ex = math_parse(expression)
		print(expression, "=", ex.evaluate_numpy(values))
	import types
	import numba

	numba.jit
	types.FunctionType()
	sys.exit(0)
	import vaex as vx
	from setuptools import setup, Extension
	ds = vx.example()
	scope = Scope(ds)
	result = eval("x + y*3", {}, scope)
	import StringIO
	f = StringIO.StringIO()
	f.write("// this is generated code =====\n")
	scope.write(f)
	f.write("values[0] = %s;" % result.c_code())
	f.write("// end of generated code =====\n")
	print(f.getvalue())
	import os
	import platform
	from distutils.sysconfig import get_python_inc

	base_path = os.path.expanduser("~/.vaex/ext")
	#build_path = os.path.expanduser("~/.vaex/ext/build")
	#build_path = os.path.expanduser("~/.vaex/ext/temp")
	if not os.path.exists(base_path):
		os.makedirs(base_path)

	filename = os.path.join(base_path, "code.cpp")
	template = file(os.path.expanduser("src/vaex/vaexfast_template.cpp")).read()
	with open(filename, "w") as file:
		file.write(template.replace("{code_nd}", f.getvalue()))
	try:
		import numpy
		numdir = os.path.dirname(numpy.__file__)
	except:
		numdir = None

	if numdir is None:
		print("numpy not found, cannot install")

	include_dirs = []
	library_dirs = []
	libraries = []
	defines = []
	if "darwin" in platform.system().lower():
		extra_compile_args = ["-mfpmath=sse", "-O3", "-funroll-loops"]
	else:
		#extra_compile_args = ["-mfpmath=sse", "-msse4", "-Ofast", "-flto", "-march=native", "-funroll-loops"]
		extra_compile_args = ["-mfpmath=sse", "-msse4", "-Ofast", "-flto", "-funroll-loops"]
		#extra_compile_args = ["-mfpmath=sse", "-O3", "-funroll-loops"]
		extra_compile_args = ["-mfpmath=sse", "-msse4a", "-O3", "-funroll-loops"]
	extra_compile_args.extend(["-std=c++0x"])

	include_dirs.append(os.path.join(get_python_inc(plat_specific=1), "numpy"))
	if numdir is not None:
		include_dirs.append(os.path.join(numdir, "core", "include"))

	ex = Extension("vaexfast_user", [filename],
                include_dirs=include_dirs,
                library_dirs=library_dirs,
                libraries=libraries,
                define_macros=defines,
                extra_compile_args=extra_compile_args
                )
	from setuptools import setup, Extension
	setup(script_args=["build_ext", "--build-lib=" + base_path], ext_modules=[ex])



	if 0:
		logger.setLevel(logging.DEBUG)

		expr1 = "log(a[:,index]) + index"
		expr1 = "log(a[index]) + index"
		expr1 = "log(a[-1]) - index + 1 + 1**2 + index**index + d(a,b) + 1.2"
		expr1 = "-2.5 * log10(SPECTROFLUX[:,2])"
		scope = Scope(None)
		result = eval(expr1, {}, scope)
		print(("expression", expr1))
		print(("should equal", repr(result)))
		print("translating")
		print()
		newexpr, vars = translate(expr1)
		print((newexpr, vars))

		expr = "xc + yc"
		newexpr, vars = translate(expr, {"xc":"x+1", "yc":"sqrt(x**2+y**2)"})
		print((newexpr, vars))


		expr = "(xc + yc)/r"
		replacements = collections.OrderedDict()
		replacements["xc"] = "x+1"
		replacements["yc"] = "y-1"
		replacements["r"] = "sqrt(xc**2+yc**2)"
		print((list(replacements.items())))
		newexpr, vars = translate(expr, replacements)
		print((newexpr, vars))







__author__ = 'maartenbreddels'
import sys
import json
import paho.mqtt.client as mqtt


def main(argv):


	type = argv[1]

	# The callback for when the client receives a CONNACK response from the server.
	def on_connect(client, userdata, flags, rc):
		print("Connected with result code "+str(rc))

		# Subscribing in on_connect() means that if we lose the connection and
		# reconnect then subscriptions will be renewed.
		if type == "listen":
			client.subscribe("vaex/#")

	# The callback for when a PUBLISH message is received from the server.
	def on_message(client, userdata, msg):
		print(msg.topic+" "+str(msg.payload))

	client = mqtt.Client()
	client.on_connect = on_connect
	client.on_message = on_message

	client.connect("broker.hivemq.com", 1883, 60)

	# Blocking call that processes network traffic, dispatches callbacks and
	# handles reconnecting.
	# Other loop*() functions are available that give a threaded interface and a
	# manual interface.
	if type == "listen":
		client.loop_forever()
	else:
		client.loop_start()
		print(client.publish("vaex/zoom", json.dumps({'x':1, 'y':2, }), 2, False))

if __name__ == "__main__":
	main(sys.argv)


# -*- coding: utf-8 -*-
import sys
import os
import platform

darwin = "darwin" in platform.system().lower()
frozen = getattr(sys, 'frozen', False)
if frozen and len(sys.argv) > 1 and sys.argv[1].startswith("-psn"): # is the 'app' is opened in osx, just stars without arguments
	dirname = os.path.join(os.path.dirname(sys.argv[0]), "..", "..", "..")
	os.chdir(dirname)


if darwin and frozen: # on newer osx versions we get a lot of broken pipes when writing to stdout
	directory = os.path.expanduser("~/.vaex")
	if not os.path.exists(directory):
		os.makedirs(directory)
	sys.stdout = open(os.path.expanduser('~/.vaex/stdout.txt'), 'w')
	sys.stderr = open(os.path.expanduser('~/.vaex/stderr.txt'), 'w')


#print darwin, platform.system()


usage = """usage veax [-h] {webserver,convert,...}

optional arguments:
  -h, --help          show this help message and exit

positional arguments:
    webserver           start the vaex webserver
    convert             convert datasets
    benchmark           run benchmarks
    meta                import/export meta data
    alias               manage aliases
    stat                print statistics/info about dataset
    test                run unittests
    ...                 anything else will start up the gui, see usage below


All other cases will start up the gui:
usage: vaex [input expression [expression ...] [key=value ...]
            [-|--|+|++] input expression...]

input        input file or url (if an url is provided, a dataset name should follow)
expression   expressions to plot, for instance "x**2", "y" "x+y"
-|--|+|++    - will open a new window, no input is required
             -- will open a new window with a new dataset (input is required)
             + will add a new layer, no input is required
             ++ will add a new layer with a new dataset (input is required)
key=value    will set properties of the plot, see a list of values below

there can be 1, 2 or 3 expressions, resulting in a 1d histogram, a 2d density plot or a
 3d volume rendering window.

key=value options:
amplitude          (string) expression for how the histogram translates to an amplitude,
                    available grids:
                       counts: N dimensional grid containing the counts
                       weighted: N dimensional grid containing the sum of
                            the weighted values
weight             (string) expression for the weight field
grid_size          (integer, power of 2), defines the size of the histogram grid, example
                       grid_size=64 grid_size=129
vector_grid_size   similar, for the vector histogram
vx, vy, vz         (string) expressions for the x,y and z component of the vector grid


Examples:
# single window, showing log(counts+1) as amplitude
$ vaex example.hdf5 x y amplitude="log(counts+1)'

# adding a second window, showing x vs z
vaex example.hdf5 x y amplitude="log(counts+1)" - x z amplitude="log(counts+1)"

# adding a second window with a new dataset
vaex example.hdf5 x y amplitude="log(counts+1)" -- example2.hdf5 x z amplitude="log(counts+1)"

# single window, with an extra layer
vaex example.hdf5 x y amplitude="log(counts+1)" + y x amplitude="log(counts+1)"

# single window, with an extra layer from a new dataset
vaex example.hdf5 x y amplitude="log(counts+1)" ++ example2.hdf5 x y amplitude="log(counts+1)"

see more examples at http://TODO

"""
def main(args=None):
	if args is None:
		args = sys.argv
	if frozen and len(args) > 1 and args[1].startswith("-psn"): # is the 'app' is opened in osx, just start without arguments
		import vaex.ui.main
		vaex.ui.main.main([])
	else:
		if len(args) > 1 and args[1] in ["-h", "--help"]:
			print(usage)
			sys.exit(0)
		if len(args) > 1 and args[1] == "version":
			import vaex.version
			if frozen:
				extra = " (build on %s using Python %s)" % (vaex.version.osname, sys.version)
			else:
				extra = " (using Python %s)" % (sys.version)
			print(vaex.__full_name__ + extra)
		elif len(args) > 1 and args[1] == "webserver":
			import vaex.webserver
			vaex.webserver.main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		elif len(args) > 1 and args[1] == "benchmark":
			import vaex.benchmark
			vaex.benchmark.main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		elif len(args) > 1 and args[1] == "convert":
			import vaex.export
			vaex.export.main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		elif len(args) > 1 and args[1] == "meta":
			import vaex.meta
			vaex.meta.main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		elif len(args) > 1 and args[1] == "alias":
			import vaex.misc_cmdline
			vaex.misc_cmdline.alias_main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		elif len(args) > 1 and args[1] == "stat":
			import vaex.misc_cmdline
			vaex.misc_cmdline.stat_main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		elif len(args) > 1 and args[1] == "test":
			import vaex.test.__main__
			vaex.test.__main__.main([os.path.basename(args[0]) +" " +args[1]]  + args[2:])
		else:
			import vaex.ui.main
			vaex.ui.main.main(args[1:])


if __name__ == "__main__":
    main()